////////////////////////////////////
// raymarching.h
// Rendering with GPU ray marching
// Original coding by Pascal Gautron, February 2007
////////////////////////////////////



#ifndef _RAYMARCHING_H_
#define _RAYMARCHING_H_
#include <volumerenderer.h>
#include <perftimer.h>
#include <matrix.h>
extern float d, theta, phi;
class CRayMarching : public CVolumeRenderer
{
public:
  CRayMarching();
  virtual ~CRayMarching() {}

  
  virtual void Render();

  virtual void Menu(int item);
  virtual void CreateMenus();

  virtual void Key(unsigned char key, int x, int y);
  virtual bool LoadData(const char *iniFile);

private:
  virtual bool LoadData(const char *baseName, const char *extension, int numberOfFiles, int digits, int start);
  void ComputeGradientsCPU(int width, int height, int depth);
  void ComputeGradientsGPU(int width, int height, int depth);
  void DisplayInfoPanel();

  float _threshold;
  int _nSteps;
  int _start;

  C3DTexture<unsigned char> _3dTex;
  GLuint _texID;
  int _depth;
  char _tmpFileName[256];
  const char *GenFileName(const char *baseName, const char *extension, int id, int digits);

  CShader *_shader;

  C3DTexture<unsigned char> _normalTex;
  //C3DTexture<float> _normalTex;
  GLuint _normalTexID;

  CPerfTimer _timer;
  int _frameCounter;
  float _fps;
  bool _infoPanel;
};


#endif //_RAYMARCHING_H_