////////////////////////////////////
// raycasting.cpp
// Raycasting-based volume renderer
// Original coding by Pascal Gautron, July 2006
////////////////////////////////////



#include <raycasting.h>

bool CRayCasting::LoadData(const char *baseName, const char *extension, int numberOfFiles, int digits)
{
  STATUS << "Loading images..." << FLUSH;
  GenFileName(baseName, extension, 1, digits);
  ilInit();
  ilImage img;
  if (!img.Load(_tmpFileName))
  {
    WARNING("Image file " << _tmpFileName << " cannot be opened");
    return false;
  }

  _depth = numberOfFiles;
  _3dTex.Init(img.Width(), img.Height(), numberOfFiles, 1);
  _3dTex.SetZSlice(0, img.GetData());

  

  for(int i=2; i <= numberOfFiles; i++)
  {
    GenFileName(baseName, extension, i, digits);
    if (!img.Load(_tmpFileName))
    {
      WARNING("Image file " << _tmpFileName << " cannot be opened");
      return false;
    }
    _3dTex.SetZSlice(i-1, img.GetData());
  }
  STATUS << "OK" << ENDL;
  STATUS << "Filling GL Texture..." << FLUSH;
  glGenTextures(1, &_texID);
  glBindTexture(GL_TEXTURE_3D, _texID);
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexImage3D(
      GL_TEXTURE_3D,
      0,
      GL_ALPHA,
      img.Width(),
      img.Height(),
      numberOfFiles,
      0,
      GL_ALPHA,
      GL_UNSIGNED_BYTE,
      _3dTex.GetData()
      );
    STATUS << "OK" << ENDL;
    


    if (_shader==NULL)
    {
      _shader = new CShader("./Shaders/test.vertex", "./Shaders/test.fragment");
    }
  return true;
}

const char *CRayCasting::GenFileName(const char *baseName, const char *extension, int id, int digits)
{
  sprintf(_tmpFileName, "%s", baseName);
  int pos = (int)strlen(baseName);
  

  int tmpid = id;
  int iddigits=0;
  while(tmpid > 0)
  {
    iddigits++;
    tmpid/= 10;
  }
  iddigits = max(1, iddigits);
  int zeroes = digits-iddigits;
  for(int i=0;i < zeroes; i++)
    _tmpFileName[pos+i] = '0';
  sprintf(_tmpFileName+pos+i, "%d.%s", id, extension);
  return _tmpFileName;
}

void CRayCasting::Render()
{

  float from[4];
  _shader->Activate();
      from[0] = d*cos(phi)*sin(theta);
    from[1] = d*sin(phi)*sin(theta);
    from[2] = d*cos(theta);
    from[3] = 1.0f;
    _shader->SetFloatUniform("from", 4, 1, from);
    _shader->SetIntUniform("tex", 0);
  glEnable(GL_DEPTH_TEST);
  glDisable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glClearColor(0.2f, 0.2f, 0.2f, 1.0f);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
  glDisable(GL_CULL_FACE);
  glColor4f(1.0f, 1.0f, 1.0f, 0.5f);
  glEnable(GL_TEXTURE_3D);
  glBindTexture(GL_TEXTURE_3D, _texID);
  
  //glPushMatrix();
glColor3f(1,1,1);
//_shader->Activate();
  DrawCube();
  /*glTranslatef(-0.5f, -0.5f, 0.5f);
  glBegin(GL_QUADS);
  for(int i=0; i < _depth; i++)
  {
    glTexCoord3f(0.0f, 0.0f, i/(float)_depth);
    glVertex3f(0.0f, 0.0f, -i/(float)_depth);

    glTexCoord3f(1.0f, 0.0f, i/(float)_depth);
    glVertex3f(1.0f, 0.0f, -i/(float)_depth);

    glTexCoord3f(1.0f, 1.0f, i/(float)_depth);
    glVertex3f(1.0f, 1.0f, -i/(float)_depth);

    glTexCoord3f(0.0f, 1.0f, i/(float)_depth);
    glVertex3f(0.0f, 1.0f, -i/(float)_depth);
  }
  glEnd();
  glPopMatrix();*/
  
}
void CRayCasting::DrawCube() const
{
  glPushMatrix();
  glTranslatef(-0.5f, -0.5f, 0.5f);
  glBegin(GL_QUADS);
  //Front
  glTexCoord3f(0,0,0);
  glVertex3f(0,0,0);
  glTexCoord3f(1,0,0);
  glVertex3f(1,0,0);
  glTexCoord3f(1,1,0);
  glVertex3f(1,1,0);
  glTexCoord3f(0,1,0);
  glVertex3f(0,1,0);

  //Back
  glTexCoord3f(0,1,1);
  glVertex3f(0,1,-1);
  glTexCoord3f(1,1,1);
  glVertex3f(1,1,-1);
  glTexCoord3f(1,0,1);
  glVertex3f(1,0,-1);
  glTexCoord3f(0,0,1);
  glVertex3f(0,0,-1);

  //Left
  glTexCoord3f(0,0,0);
  glVertex3f(0,0,0);
  glTexCoord3f(0,1,0);
  glVertex3f(0,1,0);
  glTexCoord3f(0,1,1);
  glVertex3f(0,1,-1);
  glTexCoord3f(0,0,1);
  glVertex3f(0,0,-1);

  //Right
  glTexCoord3f(1,0,1);
  glVertex3f(1,0,-1);
  glTexCoord3f(1,1,1);
  glVertex3f(1,1,-1);
  glTexCoord3f(1,1,0);
  glVertex3f(1,1,0);
  glTexCoord3f(1,0,0);
  glVertex3f(1,0,0);

  //Top
  glTexCoord3f(0,1,0);
  glVertex3f(0,1,0);
  glTexCoord3f(1,1,0);
  glVertex3f(1,1,0);
  glTexCoord3f(1,1,1);
  glVertex3f(1,1,-1);
  glTexCoord3f(0,1,1);
  glVertex3f(0,1,-1);

  //Bottom
  glTexCoord3f(0,0,1);
  glVertex3f(0,0,-1);
  glTexCoord3f(1,0,1);
  glVertex3f(1,0,-1);
  glTexCoord3f(1,0,0);
  glVertex3f(1,0,0);
  glTexCoord3f(0,0,0);
  glVertex3f(0,0,0);

  glEnd();

  glPopMatrix();


}


