////////////////////////////////////
// main.h
// Main
// Original coding by Pascal Gautron, July 2006
////////////////////////////////////

#define USESHEARWARP 1
#define USERAYMARCHING 0



#define IMGPATH "data/bone/bon"
#define IMGCOUNT 46
#define IMGDIGITS 4


#include <gllib.h>
#include <core.h>
#include <shearwarp.h>
#include <raymarching.h>
CVolumeRenderer *vrender=NULL;
float d=5.0f, theta=M_PI/2.0f, phi=M_PI;
float ltheta = 0.0f, lphi = 0.0f;

float black[] = {0.0f, 0.0f, 0.0f, 1.0f};
float white[] = {1.0f, 1.0f, 1.0f, 1.0f};
//float lpos[4] = {-1.0f,   -1.0f,   -5.0f,    1.0f};
float lpos[4] = {0.0f, 0.0f, 5.0f, 1.0f};
float lcol[4] = {1.0f, 1.0f, 1.0f,  1.0f};


void SetLight()
{
  float dst = 5.0f;
  glPushMatrix();
  glTranslatef(
    dst*cosf(lphi)*sinf(ltheta),
    dst*sinf(lphi)*sinf(ltheta),
    dst*cosf(ltheta));
    glLightfv(GL_LIGHT0, GL_DIFFUSE, lcol);
    glLightfv(GL_LIGHT0, GL_SPECULAR, lcol);
    glLightfv(GL_LIGHT0, GL_AMBIENT, black);
    glLightfv(GL_LIGHT0, GL_POSITION, lpos);
    glEnable(GL_LIGHT0);
    glPopMatrix();
}

void GLUTDisplay()
{
  glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  //gluPerspective(60.0f, 1.0f, 0.1f, 100.0f);
  glOrtho(-2.0f, 2.0f, -2.0f, 2.0f, 0.1f, 100.0f);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  

  SetLight();

  /*gluLookAt(
    d*cos(phi)*sin(theta),
    d*sin(phi)*sin(theta),
    d*cos(theta),
    0,0,0,
    0,1,0);*/
  /*gluLookAt(
      -phi,
      theta,
      5.0f,
      0,0,0,
      0,1,0);*/

  glTranslatef(0.0f, 0.0f, -5.0f);
  glRotatef(theta, 1.0f, 0.0f, 0.0f);
  glRotatef(phi, 0.0f, 0.0f, 1.0f);
	

  //glutSolidSphere(1.0f, 20, 20);

  if (vrender)
    vrender->Render();
  glUseProgramObjectARB(0);
  
  glDisable(GL_DEPTH_TEST);
  glEnable(GL_DEPTH_TEST);
  glutSwapBuffers();
}
void GLUTReshape(int width, int height)
{
  glViewport(0,0,width, height);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  //gluPerspective(60.0f, width/(float)height, 0.1f, 100.0f);
  glOrtho(-4.0f, 4.0f, -4.0f, 4.0f, 0.1f, 100.0f);
}

void GLUTKeyboard(unsigned char key, int x, int y)
{
  if (vrender)
    vrender->Key(key, x, y);
  glutPostRedisplay();
}

void GLUTSpecialKeys(int key, int x, int y)
{
  float step = 0.1f;
  switch(key)
  {
  case GLUT_KEY_DOWN:
    if (ltheta > step*0.5f)
    {
      ltheta -= step*0.5f;
    }
    break;
  case GLUT_KEY_UP:
    if (ltheta < -step*0.5f+M_PI)
    {
      ltheta += step*0.5f;
    }
    break;
  case GLUT_KEY_LEFT:
    lphi += step;
    break;
  case GLUT_KEY_RIGHT:
    lphi -= step;
    break;
  }


  glutPostRedisplay();
}

void GLUTMenu(int item)
{
  if (vrender)
    vrender->Menu(item);
}

void GLUTIdle()
{
  glutPostRedisplay();
}


int oldx, oldy;
int mouseMode = 0;
void GLUTMouse(int button, int state, int x, int y)
{
  
  if (state == GLUT_DOWN) // && button == GLUT_LEFT_BUTTON)
  {
    switch(button)
    {
    case GLUT_LEFT_BUTTON: // Move mode
      mouseMode = 0;
      break;
    case GLUT_RIGHT_BUTTON: // Unassigned
      mouseMode = 1;
      break;
    case GLUT_MIDDLE_BUTTON: // Zoom mode
      mouseMode = 2;
      break;
    }
    oldx = x; oldy = y;
  }
}

void GLUTMotion(int x, int y)
{
  switch(mouseMode)
  {
  case 0:
    phi += (oldx-x)*0.1f;
    theta += (oldy-y)*0.1f;
    break;
  case 2:
    d += (oldy-y)*0.1f;
    if (vrender)
    {
      vrender->_zoom += (oldy-y)*0.1f;
    }
    break;
  }
  oldx = x;
  oldy = y;

}




// GLUT starting function
// Retuns the window ID
int CreateGLUT(int posX, int posY, int sizeX, int sizeY, const char *title)
{
  int winID;
  // Enable depth buffer, double buffering, RGBA and accumulation buffer
  //glutInitDisplayMode(GLUT_DEPTH  | GLUT_DOUBLE | GLUT_RGBA | GLUT_ACCUM);
  glutInitDisplayMode(
    GLUT_DEPTH  |  // Depth buffer
    GLUT_DOUBLE |  // Use double buffering
    GLUT_RGBA   |  // Color is RGBA
//    GLUT_ACCUM  |  // Use accumulation buffer
    GLUT_ALPHA    // Use alpha channel (required to have real RGBA on nVidia Quadro)
//    GLUT_STENCIL  // Use stencil buffer
);
  glutInitWindowPosition(posX,posY);
  glutInitWindowSize(sizeX,sizeY);
  winID = glutCreateWindow(title);

  glutKeyboardFunc(GLUTKeyboard);
  glutSpecialFunc(GLUTSpecialKeys);
  glutReshapeFunc(GLUTReshape);
  glutDisplayFunc(GLUTDisplay);
  glutIdleFunc(GLUTIdle);
  glutMouseFunc(GLUTMouse);
  glutMotionFunc(GLUTMotion);
  //glutPassiveMotionFunc(_PassiveMotion);
  //glutVisibilityFunc(_Visibility);

  glutCreateMenu(GLUTMenu);
  glutAttachMenu(GLUT_RIGHT_BUTTON);

  GLenum err = glewInit();

  if (GLEW_OK != err)
  {
    // problem: glewInit failed, something went seriously wrong
    FATAL(glewGetErrorString(err));
  }
  return winID; 
}

int main(int argc, char **argv)
{
  CreateGLUT(0,0,512,512,"Volume Renderer");
#if USESHEARWARP
  CShearWarp *rc = new CShearWarp();
  rc->CreateMenus();
#endif
#if USERAYMARCHING
  CRayMarching *rc = new CRayMarching();
#endif
  vrender = rc;
  //rc->LoadData("C:/Pascal/Code/VolumeRender/data/bone/bon", "BMP", 46, 4);
  if (argc > 1)
    rc->LoadData(argv[1]);
  else
    rc->LoadData("volumerender.ini");
  glutMainLoop();
  return 0;
}