////////////////////////////////////
// raycasting.cpp
// Raycasting-based volume renderer
// Original coding by Pascal Gautron, July 2006
////////////////////////////////////



#include <shearwarp.h>
using namespace std;


CShearWarp::CShearWarp()
:_shader(NULL), _sampling(5), _threshold(0.5f), _frameCounter(0), _fps(0.0f), _infoPanel(true), _start(0)
  {
    int viewport[4];
    glGetIntegerv(GL_VIEWPORT, viewport);
    _fbo = new CFBO(viewport[2], viewport[3], 1, true);
    _fbo->Bind();
    glBindTexture(GL_TEXTURE_2D, _fbo->GetTexture(0));
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, viewport[2], viewport[3], 0, GL_RGBA, GL_FLOAT, NULL);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glBindTexture(GL_TEXTURE_2D, 0);

    _fbo->AttachTexture2D(0);
    _fbo->CheckStatus();
    glReadBuffer(GL_NONE);
    _fbo->CheckStatus();
    glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);
    _fbo->CheckStatus();
    _fbo->UnBind();
    Matrix4x4::Identity(_curTexMatrix);
    strcpy(slicing, "Slicing: XY");

  }


static void ReadLine(char *buffer, FILE *file)
{
  int i=0;
  char tmp=' ';
  while(fscanf(file, "%c", &tmp) != EOF && tmp != '\n')
  {
    buffer[i++] = tmp;
  }
  buffer[i] = '\0';
}

bool CShearWarp::LoadData(const char *iniFile)
{
  FILE *file = fopen(iniFile, "r");
  if (!file)
  {
    STATUS << "Configuration file " << iniFile << "not found";
    return false;
  }
  char fileBase[1024];
  char ext[4];
  char fileCount[10];
  char digits[10];
  char start[10];
  ReadLine(fileBase, file);
  ReadLine(ext, file);
  ReadLine(fileCount, file);
  ReadLine(digits, file);
  ReadLine(start, file);
  
  return LoadData(fileBase, ext, atoi(fileCount), atoi(digits), atoi(start)-1);

}

bool CShearWarp::LoadData(const char *baseName, const char *extension, int numberOfFiles, int digits, int start)
{
  STATUS << "Loading images..." << FLUSH;
  GenFileName(baseName, extension, 1+start, digits);
  ilInit();
  ilImage img;
  if (!img.Load(_tmpFileName))
  {
    WARNING("Image file " << _tmpFileName << " cannot be opened");
    return false;
  }



  int width = img.Width(), height = img.Height(), depth = numberOfFiles;
  _depth = numberOfFiles;
  _3dTex.Init(img.Width(), img.Height(), numberOfFiles, 1);

  _3dTex.SetZSlice(0, img.GetData());

  

  for(int i=2; i <= numberOfFiles; i++)
  {
    GenFileName(baseName, extension, i+start, digits);
    if (!img.Load(_tmpFileName))
    {
      WARNING("Image file " << _tmpFileName << " cannot be opened");
      //return false;
      continue;
    }
    _3dTex.SetZSlice(i-1, img.GetData());
  }
  STATUS << "OK" << ENDL;
  STATUS << "Filling GL Texture..." << FLUSH;
  glGenTextures(1, &_texID);
  glBindTexture(GL_TEXTURE_3D, _texID);
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexImage3D(
      GL_TEXTURE_3D,
      0,
      GL_ALPHA,
      img.Width(),
      img.Height(),
      numberOfFiles,
      0,
      GL_ALPHA,
      GL_UNSIGNED_BYTE,
      _3dTex.GetData()
      );
    STATUS << "OK" << ENDL;
 
  ComputeGradientsCPU(width, height, depth);




    STATUS << "Compiling shear shader..." << FLUSH;
    if (_shader==NULL)
    {
      _shader = new CShader("./Shaders/shearwarp.vertex", "./Shaders/shearwarp.fragment");
      _shader->Activate();
      _shader->SetFloatUniform("depth", _depth);
      _shader->SetFloatUniform("step", 1.0f/_depth);
      _shader->SetIntUniform("tex", 0);
      _shader->SetIntUniform("normalTex", 1);
    }
    STATUS << "OK" << ENDL;
  return true;
}

void CShearWarp::ComputeGradientsCPU(int width, int height, int depth)
{
    STATUS << "Computing gradients... " << FLUSH;
    CPerfTimer gradTime(TRUE);
  _normalTex.Init(width, height, depth, 3);

  unsigned char normalUB[3];
  int distance = 1;
  float weight = 1.0f/(2.0f*distance*255.0f);
    for(int z = distance; z < depth-distance; z++)
    {
      for(int y=distance; y < height-distance; y++)
      {
        for(int x=distance; x < width-distance; x++)
        {
          float normal[3] = {0.0f, 0.0f, 0.0f};
          for(int d=-distance; d < 0; d++)
          {
            unsigned char *xOp = _3dTex.GetPixel(x+d, y, z);
            unsigned char *yOp = _3dTex.GetPixel(x, y+d, z);
            unsigned char *zOp = _3dTex.GetPixel(x, y, z+d);
            normal[0] += -weight*(*xOp);
            normal[1] += -weight*(*yOp);
            normal[2] += -weight*(*zOp);
          }
          for(int d=1; d <= distance; d++)
          {
            unsigned char *xOp = _3dTex.GetPixel(x+d, y, z);
            unsigned char *yOp = _3dTex.GetPixel(x, y+d, z);
            unsigned char *zOp = _3dTex.GetPixel(x, y, z+d);
            normal[0] += weight*(*xOp);
            normal[1] += weight*(*yOp);
            normal[2] += weight*(*zOp);
          }

          float len = InvSqrt(
            normal[0]*normal[0]+
            normal[1]*normal[1]+
            normal[2]*normal[2]);
          normal[0]*= len;
          normal[1]*= len;
          normal[2]*= len;

          normal[0] = (normal[0]+1.0f)*0.5f;
          normal[1] = (normal[1]+1.0f)*0.5f;
          normal[2] = (normal[2]+1.0f)*0.5f;

          normalUB[0] = (unsigned char)(255.0f*normal[0]);
          normalUB[1] = (unsigned char)(255.0f*normal[1]);
          normalUB[2] = (unsigned char)(255.0f*normal[2]);

          _normalTex.SetPixel(x, y, z, normalUB);
        }
      }
    }

  for(int z=0; z < distance; z++)
  {
     for(int y=0; y < height; y++)
      {
        for(int x=0; x < width; x++)
        {
          memcpy(_normalTex.GetPixel(x, y, z), _normalTex.GetPixel(x, y, distance), 3*sizeof(unsigned char)); 
          memcpy(_normalTex.GetPixel(x, y, depth-z-1), _normalTex.GetPixel(x, y, depth-distance-1), 3*sizeof(unsigned char)); 
        }
      }
  }

  for(int z=0; z < depth; z++)
  {
     for(int y=0; y < distance; y++)
      {
        for(int x=0; x < width; x++)
        {
          memcpy(_normalTex.GetPixel(x, y, z), _normalTex.GetPixel(x, distance, z), 3*sizeof(unsigned char)); 
          memcpy(_normalTex.GetPixel(x, height-y-1, z), _normalTex.GetPixel(x, height-distance-1, z), 3*sizeof(unsigned char)); 
        }
      }
  }
  for(int z=0; z < depth; z++)
  {
     for(int y=0; y < height; y++)
      {
        for(int x=0; x < distance; x++)
        {
          memcpy(_normalTex.GetPixel(x, y, z), _normalTex.GetPixel(distance, y, z), 3*sizeof(unsigned char)); 
          memcpy(_normalTex.GetPixel(width-x-1, y, z), _normalTex.GetPixel(width-distance-1, y, z), 3*sizeof(unsigned char)); 
        }
      }
  }

  gradTime.Stop();

  glGenTextures(1, &_normalTexID);
  glBindTexture(GL_TEXTURE_3D, _normalTexID);
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexImage3D(
      GL_TEXTURE_3D,
      0,
      GL_RGB,
      width,
      height,
      depth,
      0,
      GL_RGB,
      GL_UNSIGNED_BYTE,
      _normalTex.GetData()
      );

    STATUS << "OK (" << gradTime.Elapsedms() << "ms)" << ENDL;
}

void CShearWarp::ComputeGradientsGPU(int width, int height, int depth)
{
      STATUS << "Computing gradients... " << FLUSH;
    CPerfTimer gradTime(TRUE);
  CFBO gfbo(width, height, 1, false);

  CShader gradShader("./Shaders/gradient.vertex", "./Shaders/gradient.fragment");

  gradShader.Activate();
  float step[] = {1.0f/width, 1.0f/height, 1.0f/depth};
  
  gradShader.SetFloatUniform("step",  3, 1, step);
  gradShader.SetIntUniform("tex", 0);
  

  gfbo.Bind();
  

  glGenTextures(1, &_normalTexID);
  glBindTexture(GL_TEXTURE_3D, _normalTexID);
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexImage3D(
      GL_TEXTURE_3D,
      0,
      GL_RGB,
      width,
      height,
      depth,
      0,
      GL_RGB,
      GL_UNSIGNED_BYTE,
      NULL
      );


    glBindTexture(GL_TEXTURE_3D, 0);

  glBindTexture(GL_TEXTURE_2D, gfbo.GetTexture(0));
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexImage2D(
      GL_TEXTURE_2D,
      0,
      GL_RGB,
      width,
      height,
      0,
      GL_RGB,
      GL_UNSIGNED_BYTE,
      NULL
      );
    glBindTexture(GL_TEXTURE_2D, 0);

  
  glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    for(int i=0; i < depth; i++)
    {
      //glFramebufferTexture3DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_3D, _normalTexID, 0, i);
      //gfbo.AttachTexture3D(0, _normalTexID, i);
      gfbo.AttachTexture2D(0);
      gfbo.CheckStatus();
      glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);
      glReadBuffer(GL_COLOR_ATTACHMENT0_EXT);
      glActiveTexture(GL_TEXTURE0);
      glBindTexture(GL_TEXTURE_3D, _texID);
      glBegin(GL_QUADS);
      glTexCoord3f(0.0f, 0.0f, 1.0f-(i)/(float)depth);
      //glTexCoord3f(0.0f, 0.0f, 0.5f);
      glVertex2f(-1.0f, -1.0f);

      glTexCoord3f(1.0f, 0.0f, 1.0f-(i)/(float)depth);
      //glTexCoord3f(1.0f, 0.0f, 0.5f);
      glVertex2f(1.0f, -1.0f);

      glTexCoord3f(1.0f, 1.0f, 1.0f-(i)/(float)depth);
      //glTexCoord3f(1.0f, 1.0f, 0.5f);
      glVertex2f(1.0f, 1.0f);

      glTexCoord3f(0.0f, 1.0f, 1.0f-(i)/(float)depth);
      //glTexCoord3f(0.0f, 1.0f, 0.5f);
      glVertex2f(-1.0f, 1.0f);
      glEnd();
      glBindTexture(GL_TEXTURE_3D, _normalTexID);
      //glCopyTexSubImage3D(GL_TEXTURE_3D, 0, 0, 0, i, 0, 0, width, height);
      glCopyTexSubImage3D(GL_TEXTURE_3D, 0, 0, 0, i, 0, 0, width, height);
  
    }
    STATUS << "OK (" << gradTime.Elapsedms() << "ms)" << ENDL;
    DisplayGLError();
    gfbo.UnBind();
}

const char *CShearWarp::GenFileName(const char *baseName, const char *extension, int id, int digits)
{
  sprintf(_tmpFileName, "%s", baseName);
  int pos = (int)strlen(baseName);
  

  int tmpid = id;
  int iddigits=0;
  while(tmpid > 0)
  {
    iddigits++;
    tmpid/= 10;
  }
  iddigits = max(1, iddigits);
  int zeroes = digits-iddigits;
  for(int i=0;i < zeroes; i++)
    _tmpFileName[pos+i] = '0';
  sprintf(_tmpFileName+pos+i, "%d.%s", id, extension);
  return _tmpFileName;
}

void CShearWarp::Render()
{

  _timer.Start();

  //_fbo->UnBind();
  _fbo->Bind();




  _fbo->SetDrawBuffer(0);
  glDisable(GL_DEPTH_TEST);
  glDepthMask(GL_FALSE);
  glEnable(GL_BLEND);
  glDisable(GL_CULL_FACE);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  //glBlendFunc(GL_SRC_ALPHA, GL_DST_ALPHA);
  //glBlendFunc(GL_ZERO, GL_DST_ALPHA);
  glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
  glDisable(GL_CULL_FACE);
  glColor4f(1.0f, 1.0f, 1.0f, 0.5f);
  glEnable(GL_TEXTURE_3D);


  glActiveTexture(GL_TEXTURE1);
  glEnable(GL_TEXTURE_3D);
  glBindTexture(GL_TEXTURE_3D, _normalTexID);

  glActiveTexture(GL_TEXTURE0);
  glEnable(GL_TEXTURE_3D);
  glBindTexture(GL_TEXTURE_3D, _texID);

  glColor3f(0.2f,0.2f,0.2f);

  float mdlview[16];
  float proj[16];
  glGetFloatv(GL_MODELVIEW_MATRIX, mdlview);
  glGetFloatv(GL_PROJECTION_MATRIX, proj);

  float from[3];
  float zeroV[] = {0.0, 0.0, 0.0};
  float imdl[16];
  Matrix4x4::Invert(mdlview, imdl);
  Matrix4x4::MultCol(imdl, zeroV, from);

  float mview[16];
  Matrix4x4::Product(mdlview, proj, mview);

  float v0[3];
  v0[0] = 
    Matrix4x4::Elem1(mview, 1, 2)*Matrix4x4::Elem1(mview, 2, 3)-
    Matrix4x4::Elem1(mview, 2, 2)*Matrix4x4::Elem1(mview, 1, 3);
  v0[1] = 
    Matrix4x4::Elem1(mview, 2, 1)*Matrix4x4::Elem1(mview, 1, 3)-
    Matrix4x4::Elem1(mview, 1, 1)*Matrix4x4::Elem1(mview, 2, 3);
  v0[2] = 
    Matrix4x4::Elem1(mview, 1, 1)*Matrix4x4::Elem1(mview, 2, 2)-
    Matrix4x4::Elem1(mview, 2, 1)*Matrix4x4::Elem1(mview, 1, 2);

  float sx = -v0[0]/v0[2];
  float sy = -v0[1]/v0[2];

  float mshear3d[16];
  Matrix4x4::Identity(mshear3d);
  Matrix4x4::Elem1(mshear3d, 1, 3) = -sx;
  Matrix4x4::Elem1(mshear3d, 2, 3) = -sy;



  glMatrixMode(GL_TEXTURE);
//  glLoadIdentity();
  glLoadMatrixf(_curTexMatrix);



  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glLoadMatrixf(mshear3d);
  //glScalef(1.0f, 1.0f, 0.5f);
  float mtmp[16];
  Matrix4x4::Identity(mtmp);
  Matrix4x4::Elem1(mtmp, 1, 3) = sx;
  Matrix4x4::Elem1(mtmp, 2, 3) = sy;

  float mwarp[16];
  Matrix4x4::Product(mview, mtmp, mwarp);
  mwarp[3] = mwarp[7] = mwarp[11] = mwarp[12] = mwarp[13] = mwarp[14] = 0.0f;
  mwarp[14] = -5.0f;
  mwarp[15] = 1.0f;

  

  //glMatrixMode(GL_PROJECTION);
  //glLoadMatrixf(mwarp);
  _shader->Activate();
  _shader->SetFloatUniform("threshold", _threshold);
  _shader->SetFloatUniform("from", 3, 1 ,from);
  DrawShearsXY();
  glFinish();
  _fbo->UnBind();




  _shader->UseProgram(0);
  glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);







glDisable(GL_CULL_FACE);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(-1.0f, 1.0f, -1.0f, 1.0f, 0.1f, 1000.0f);
  glMatrixMode(GL_TEXTURE);
  glLoadIdentity();
glMatrixMode(GL_MODELVIEW);
  glLoadMatrixf(mwarp);

  glDisable(GL_BLEND);
  glDisable(GL_LIGHTING);
  

  glActiveTexture(GL_TEXTURE1);
  glDisable(GL_TEXTURE_3D);
  glDisable(GL_TEXTURE_2D);


  glActiveTexture(GL_TEXTURE0);
  glDisable(GL_TEXTURE_3D);
  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D, _fbo->GetTexture(0));
  glPushMatrix();
  glScalef(4.0f, 4.0f, 4.0f);
  glBegin(GL_QUADS);
  glColor3f(1.0f, 1.0f, 1.0f);
  glTexCoord2f(0.0f, 0.0f);
  glVertex3f(-1.0f, -1.0f, 0.0f);

  glTexCoord2f(1.0f, 0.0f);
  glVertex3f(1.0f, -1.0f, 0.0f);

  glTexCoord2f(1.0f, 1.0f);
  glVertex3f(1.0f, 1.0f, 0.0f);

  glTexCoord2f(0.0f, 1.0f);
  glVertex3f(-1.0f, 1.0f, 0.0f);
  glEnd();
  glPopMatrix();
  glBindTexture(GL_TEXTURE_2D, 0);


  _timer.Stop();
  _frameCounter++;

  if (_timer.Elapsed() >= 1.0)
  {
    _fps = _frameCounter / _timer.Elapsed();
    _frameCounter = 0;
    _timer -= _timer;
  }

  DisplayInfoPanel();
  

  //glutSolidSphere(5.0f, 20, 20);
}

void CShearWarp::DisplayInfoPanel()
{
if (_infoPanel)
  {
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glDisable(GL_DEPTH_TEST);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    glColor4f(0.0f, 0.75f, 0.0f, 0.7f);

    glLineWidth(3.0f);
    
    glPolygonMode(GL_FRONT, GL_FILL);
    glBegin(GL_QUADS);
    glVertex2f(-0.95f, -0.60f);
    glVertex2f(-0.95f, -0.98f);
    glVertex2f(-0.2f, -0.98f);
    glVertex2f(-0.2f, -0.60f);
    glEnd();

    glColor4f(0.0f, 0.0f, 0.0f, 1.0f);
    glPolygonMode(GL_FRONT, GL_LINE);
    glBegin(GL_QUADS);
    glVertex2f(-0.95f, -0.60f);
    glVertex2f(-0.95f, -0.98f);
    glVertex2f(-0.2f, -0.98f);
    glVertex2f(-0.2f, -0.60f);
    glEnd();
    glPolygonMode(GL_FRONT, GL_FILL);

    glColor3f(1.0f, 1.0f, 1.0f);
    float position[] = {-0.92f, -0.65f};


    glDisable(GL_BLEND);

    BitmapOutput2D(position, slicing);

    char buffer[255];
    sprintf(buffer, "Opacity (A/Q): %3.2f", _threshold);
    position[1] -= 0.05f; 
    BitmapOutput2D(position, buffer);

    position[1] -= 0.05f;
    sprintf(buffer, "Sampling (Z/S): %d", _sampling);
    BitmapOutput2D(position, buffer);

    position[1] -= 0.05f;
    sprintf(buffer, "Start slice (E/D): %d", _start);
    BitmapOutput2D(position, buffer);

    position[1] -= 0.05f;
    sprintf(buffer, "FPS: %3.2f", _fps);
    BitmapOutput2D(position, buffer);

    position[1] -= 0.05f;
    sprintf(buffer, "Arrow keys: Move light");
    BitmapOutput2D(position, buffer);

    position[1] -= 0.05f;
    sprintf(buffer, "Space: Toogle info");
    BitmapOutput2D(position, buffer);
  }
}


void CShearWarp::DrawShearsXY() const
{
  
  float increment = 1.0f/(_depth*_sampling);
  float scale = _zoom;
  glPushMatrix();
  glScalef(scale, scale, scale);
  //glTranslatef(-0.5f, -0.5f, 0.5f);
  glBegin(GL_QUADS);
  

  int start = _start*_sampling;
  int end = _sampling*_depth;

  float z = 1.0f;

  for(int i=start; i < end /*_sampling*_depth*/; i++, z-=increment)
  {
    float slice = (1.0f-z)/(_sampling*_depth);
    glTexCoord4f(0.0f, 0.0f, z, slice);
    glVertex3f(-1.0f, -1.0f, -z);

    glTexCoord4f(1.0f, 0.0f, z, slice);
    glVertex3f(1.0f, -1.0f, -z);

    glTexCoord4f(1.0f, 1.0f, z, slice);
    glVertex3f(1.0f, 1.0f, -z);

    glTexCoord4f(0.0f, 1.0f, z, slice);
    glVertex3f(-1.0f, 1.0f, -z);
  }
  glEnd();
  glPopMatrix();
}

void CShearWarp::Menu(int item)
{
  switch(item)
  {
  case 0:
    Matrix4x4::Identity(_curTexMatrix);
    //glutSetWindowTitle("Volume Renderer: ShearWarp (XY Slicing)");
    strcpy(slicing, "Slicing: XY");
    break;
  case 1:
    Matrix4x4::Zero(_curTexMatrix);
    _curTexMatrix[0] = -1.0f;
    _curTexMatrix[5] = -1.0f;
    _curTexMatrix[10] = -1.0f;
    _curTexMatrix[15] = 1.0f;
    _curTexMatrix[12] = 1.0f;
    _curTexMatrix[13] = 1.0f;
    _curTexMatrix[14] = 1.0f;
    //glutSetWindowTitle("Volume Renderer: ShearWarp (-XY Slicing)");
    strcpy(slicing, "Slicing: -XY");
    break;
  case 2:
    Matrix4x4::Zero(_curTexMatrix);
    _curTexMatrix[0] = 1.0f;
    _curTexMatrix[9] = 1.0f;
    _curTexMatrix[6] = 1.0f;

    _curTexMatrix[15] = 1.0f;
    //glutSetWindowTitle("Volume Renderer: ShearWarp (YZ Slicing)");
    strcpy(slicing, "Slicing: YZ");
    break;

  case 3:
    Matrix4x4::Zero(_curTexMatrix);
    _curTexMatrix[0] = -1.0f;
    _curTexMatrix[9] = -1.0f;
    _curTexMatrix[6] = 1.0f;

    _curTexMatrix[12] = 1.0f;
    _curTexMatrix[13] = 1.0f;
    //_curTexMatrix[14] = 1.0f;

    _curTexMatrix[15] = 1.0f;
    //glutSetWindowTitle("Volume Renderer: ShearWarp (-YZ Slicing)");
    strcpy(slicing, "Slicing: -YZ");
    break;
  case 4:
    Matrix4x4::Zero(_curTexMatrix);
    _curTexMatrix[8] = 1.0f;
    _curTexMatrix[1] = 1.0f;
    _curTexMatrix[6] = 1.0f;

    _curTexMatrix[15] = 1.0f;
    //glutSetWindowTitle("Volume Renderer: ShearWarp (ZX Slicing)");
    strcpy(slicing, "Slicing: ZX");
    break;

  case 5:
    Matrix4x4::Zero(_curTexMatrix);
    _curTexMatrix[8] = -1.0f;
    _curTexMatrix[1] = -1.0f;
    _curTexMatrix[6] = 1.0f;

    _curTexMatrix[12] = 1.0f;
    _curTexMatrix[13] = 1.0f;
    //_curTexMatrix[14] = 1.0f;

    _curTexMatrix[15] = 1.0f;


    //glutSetWindowTitle("Volume Renderer: ShearWarp (-ZX Slicing)");
    strcpy(slicing, "Slicing: -ZX");
    break;


  }
  glutPostRedisplay();
}
void CShearWarp::CreateMenus()
{
  glutAddMenuEntry("Slice XY",0);
  glutAddMenuEntry("Slice -XY",1);
  glutAddMenuEntry("Slice YZ",2);
  glutAddMenuEntry("Slice -YZ",3);
  glutAddMenuEntry("Slice ZX",4);
  glutAddMenuEntry("Slice -ZX",5);
}

void CShearWarp::Key(unsigned char key, int x, int y)
  {
    switch (key)
    {
    case 'a':
    case 'A':
      if (_threshold < 1.0f)
      {
        _threshold += 0.01f;
      }
      break;
    case 'q':
    case 'Q':
      if (_threshold > 0.01f)
      {
        _threshold -= 0.01f;
      }
      break;
    case 'Z':
    case 'z':
      _sampling++;
      break;
    case 'S':
    case 's':
      if (_sampling > 1) _sampling--;
      break;
    case ' ':
      _infoPanel = !_infoPanel;
    
    case 'E':
    case 'e':
      if (_start < _depth-1)
      {
        _start++;
        
      }
      break;
    case 'D':
    case 'd':
      if (_start > 0 )
      {
        _start--;
      }
      break;
    }
  }
